"""
This module defines the Building class, which represents buildings containing different units and their data
"""
import re
from typing import Dict, List, Optional, Union

import pandas as pd

from . import unit_graph
from .api import AccessDeniedError, api_get
from .building_unit import BaseBuildingUnit, populate_units
from .data import load_model_data
from .helpers import convenience_result_list_shortener, sanitise_datetime_input
from .price import SupplyPoint
from .weather import Weather
from .zone import Zone, populate_zones


class Building(object):
    """Building defines a building in the PreHEAT sense"""

    def __init__(self, location_id: int):
        self.location_id = location_id

        # Dict containing the details of the location
        self.location = {}
        # List of zones within the building (of class PreHEAT_API.Zone)
        self.zones = []
        # Dict w.units within the building (of class PreHEAT_API.Unit)
        self.units = {}
        # Dict w. weather information for the location (PreHEAT_API.Weather)
        self.weather = None  # type: Weather
        # Area of the building
        self.area = None
        # All characteristics get summarised in this dictionary
        # (duplicate with some attributes do to backwards-compatibility requirement)
        self.characteristics = {}

        # Supply points (for pricing)
        self.__supply_points = []

        # Load from API
        self.__populate()

        # Construct graph
        self.G = unit_graph.generate_unit_graph(self)

        # Traverse and map sub-unit references
        for node, data in self.G.nodes.data():
            # Add sub-units as new nodes
            if hasattr(data["unit"], "_related_sub_units_refs"):
                for ref in data["unit"]._related_sub_units_refs:
                    data["unit"].add_sub_unit(self.G.nodes[ref]["unit"])
            if hasattr(data["unit"], "_related_meters_refs"):
                for ref in data["unit"]._related_meters_refs:
                    data["unit"]._meters.append(self.G.nodes[ref]["unit"])

    def get_unit_graph(self):
        """

        :return:
        :rtype:
        """
        return self.G

    def load_data(
        self,
        start_date,
        end_date,
        time_resolution: str = "hour",
        components: Optional[Dict] = {},
    ) -> None:
        """

        :param start_date:
        :type start_date:
        :param end_date:
        :type end_date:
        :param time_resolution:
        :type time_resolution:
        :param components:
        :type components:
        :return:
        :rtype:
        """
        start_date = sanitise_datetime_input(start_date)
        end_date = sanitise_datetime_input(end_date)
        if components is None:
            components = {}

        if len(components) == 0:
            # Load all data on building, if nothing is specified
            self.weather.load_data(
                start_date, end_date, time_resolution, components.get("weather")
            )
            for node, data in self.G.nodes.data():
                u_i = data["unit"]
                c_i = components.get(u_i.unit_type)
                u_i.load_data(start_date, end_date, time_resolution, components=c_i)

        else:
            # Otherwise, just load specific components
            for i in components.keys():
                c_i = components.get(i)
                if i == "weather":
                    self.weather.load_data(
                        start_date, end_date, time_resolution, components=c_i
                    )
                else:
                    for u_i in self.query_units(unit_type=i):
                        u_i.load_data(
                            start_date, end_date, time_resolution, components=c_i
                        )

    def load_dataset(
        self,
        component_map,
        start_date,
        end_date,
        time_resolution: str = "hour",
        load_weather: bool = True,
    ) -> pd.DataFrame:
        """

        :param component_map:
        :type component_map:
        :param start_date:
        :type start_date:
        :param end_date:
        :type end_date:
        :param time_resolution:
        :type time_resolution:
        :param load_weather:
        :type load_weather:
        :return:
        :rtype:
        """

        df = load_model_data(component_map, start_date, end_date, time_resolution)

        if load_weather is True:
            self.weather.load_data(start_date, end_date, time_resolution)
            df_weather = self.weather.data
            df_weather.columns = "weather." + df_weather.columns
            df = pd.concat((df_weather, df), axis=1)

        return df

    def clear_data(self) -> None:
        """

        :return:
        :rtype:
        """
        self.weather.clear_data()

        for node, data in self.G.nodes.data():
            data["unit"].clear_data()

    def get_all_component_ids(self):
        """

        :return:
        :rtype:
        """
        return {
            k: v
            for node, data in self.G.nodes.data()
            for k, v in data["unit"].get_all_component_ids(True).items()
        }

    def get_all_component_details(self):
        """

        :return:
        :rtype:
        """
        all_comps = []
        for node, data in self.G.nodes.data():
            all_comps += data["unit"].get_all_component_details(prefix=True)

        return all_comps

    def query_units(
        self, unit_type: str = None, name: str = None, unit_id: int = None
    ) -> List[BaseBuildingUnit]:
        """

        :param unit_type:
        :type unit_type:
        :param name:
        :type name:
        :param unit_id:
        :type unit_id:
        :return:
        :rtype:
        """
        # If we pass a unit ID, find specific unit
        if unit_id:
            if unit_id in self.G.nodes.keys():
                return [self.G.nodes[unit_id]["unit"]]
            else:
                return []

        # If we pass a unit_type, check if we do regex search or strict search
        if unit_type:
            if unit_type[0] == "?":
                r_type = re.compile(unit_type[1:])
                type_match = lambda t: r_type.search(t)
            else:
                type_match = lambda t: t == unit_type

        # If we pass name, check if we do regex search or strict search
        if name:
            # Check if we do strict search or search by regex:
            if name[0] == "?":
                r_name = re.compile(name[1:])
                name_match = lambda n: r_name.search(n)
            else:
                name_match = lambda n: n == name

        try:
            # If we pass unit_type and name
            if unit_type and name:
                result = [
                    data["unit"]
                    for node, data in self.G.nodes.data()
                    if name_match(data["unit"].name)
                    and type_match(data["unit"].unit_type)
                ]

            elif name:
                result = [
                    data["unit"]
                    for node, data in self.G.nodes.data()
                    if name_match(data["unit"].name)
                ]

            elif unit_type:
                result = [
                    data["unit"]
                    for node, data in self.G.nodes.data()
                    if type_match(data["unit"].unit_type)
                ]

            return result

        except:
            return []

    def qu(self, *args, **kwargs) -> Union[BaseBuildingUnit, List[BaseBuildingUnit]]:
        """

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        result = self.query_units(*args, **kwargs)
        return convenience_result_list_shortener(result)

    def __populate(self) -> None:
        # Request PreHEAT API for location
        resp = api_get(f"locations/{self.location_id}", out="json")

        # Check if we have building data and if we do, extract and populate
        if "building" in resp:
            data = resp["building"]

            self.location = data["location"]
            self.area = data["buildingArea"]
            self.type = data["buildingType"]
            self.characteristics = {
                "area": self.area,
                "type": self.type,
                "apartments": data["apartments"],
            }

            self.weather = Weather(self.location_id, data["weatherForecast"])

            self.zones = populate_zones(data["zones"])

            # Order is important for dependency
            unit_types = [
                "main",
                "coldWater",
                "heating",
                "hotWater",
                "cooling",
                "electricity",
                "ventilation",
                "heatPumps",
                "indoorClimate",
                "custom",
                "localWeatherStations",
            ]

            for unit_type in unit_types:
                self.units[unit_type] = populate_units(unit_type, data[unit_type], self)

            for sp_i in data["supplyPoints"]:
                self.__supply_points.append(SupplyPoint(sp_i))

        else:
            raise AccessDeniedError("Access denied for given building")

    def get_supply_points(self):
        """

        :return:
        :rtype:
        """
        return self.__supply_points

    def __repr__(self):
        return f"""{type(self).__name__}({self.location_id}): {self.location["address"]} - {self.type}"""

    def get_zones(self, zone_ids) -> List[Zone]:
        """

        :param zone_ids:
        :type zone_ids:
        :return:
        :rtype:
        """
        res = []
        for z_i in self.zones:
            if z_i.id in zone_ids:
                res.append(z_i)
            res += z_i.get_sub_zones(zone_ids=zone_ids)
        return res

    def get_zone(self, id: int) -> Zone:
        """

        :param id:
        :type id:
        :return:
        :rtype:
        """
        zs = self.get_zones([id])
        n_zs = len(zs)
        if n_zs < 1:
            raise Exception(f"Zone not found (id={id})")
        elif n_zs > 1:
            raise Exception(f"Too many zones found for id (id={id} / {n_zs} found)")
        return zs[0]


def available_buildings() -> pd.DataFrame:
    """
    Lists available buildings

    :return: dataframe of buildings available
    :rtype:
    """
    out = api_get("locations")
    return pd.DataFrame(out["locations"])



import pandas as pd
from pandas import DataFrame

from .component import Component
from .data import load_weather_data
from .helpers import now, sanitise_datetime_input, timedelta
from .unit import Unit


class Weather(Unit):
    """Weather; an extension of Unit to handle Weather extraction"""

    def __init__(self, location_id, weather_data):
        super().__init__("weather", weather_data)
        self.location_id = location_id
        self.id = weather_data.pop("gridId", None)
        self.name = "WeatherForecast"

        for type_i in weather_data["types"]:
            self.components.append(Component(type_i))

    def load_data(
        self, start_date, end_date, time_resolution="hour", components=None
    ) -> None:
        # Parse strings to datetime object
        start_date = sanitise_datetime_input(start_date)
        end_date = sanitise_datetime_input(end_date)

        self._warn_if_data_is_loaded()
        components_to_load = self.get_all_component_ids(components=components)
        if components is not None:
            components_to_load = {
                k: components_to_load[k]
                for k in components_to_load
                if components_to_load[k] in components
            }

        if (end_date - start_date) > pd.Timedelta("180d"):
            # Batch in 6M intervals (max we can load at a time)
            parts = list(pd.date_range(start_date, end_date, freq="2Q"))

            if start_date != parts[0]:
                parts.insert(0, start_date)
            if end_date != parts[-1]:
                parts.append(end_date)

            parts[0] -= pd.Timedelta("1d")  # we add back one day later

            pairs = zip(map(lambda d: d + pd.Timedelta("1d"), parts[:-1]), parts[1:])

            dfs = []
            for start_date, end_date in pairs:
                dfs.append(
                    load_weather_data(
                        self.location_id,
                        components_to_load,
                        start_date,
                        end_date,
                        time_resolution,
                    )
                )

            res = pd.concat(dfs)

        else:
            res = load_weather_data(
                self.location_id,
                components_to_load,
                start_date,
                end_date,
                time_resolution,
            )

        self.data = res  # type: DataFrame
        self._ensure_continuity_of_data(time_resolution)

    def load_state(self, seconds_back: int = 3600, t_now=None) -> None:
        t_now = now() if t_now is None else sanitise_datetime_input(t_now)

        components_to_load = self.get_all_component_ids()
        self._state = load_weather_data(
            self.location_id,
            components_to_load,
            t_now - timedelta(seconds=seconds_back),
            t_now,
            "hour",
        )

    def get_state(
        self,
        update: bool = False,
        estimate: str = "last",
        seconds_back: int = 3600,
        **kwargs
    ) -> pd.Series:
        return super().get_state(
            update=update, estimate=estimate, seconds_back=seconds_back, **kwargs
        )



import numpy as np
import pandas as pd
from requests.models import Response

from .building_unit import BaseBuildingUnit
from .logging import Logging
from .setpoints import get_setpoint_schedule, send_setpoint_schedule


class ControlUnit(BaseBuildingUnit):
    """Control Unit; an extension of Unit to handle controls"""

    def __init__(self, unit_data, building_ref=None):
        super().__init__("control", unit_data, building_ref)
        if "active" in unit_data.keys():
            self.active = unit_data["active"]
        else:
            self.active = False

    def request_schedule(self, schedule_df) -> Response:
        if self.active is False:
            Logging().warning(
                """Warning: you are trying to control an unit that is not activated 
                (id={} / details: [unit: {} / building: [{}] {}])""".format(
                    self.id,
                    self.name,
                    self.building.location["locationId"],
                    self.building.location["address"],
                )
            )
        if schedule_df["value"].isnull().any():
            raise ValueError("requested schedule has missing values")
        if np.isinf(schedule_df["value"]).any():
            raise ValueError("requested schedule has infinite values")
        return send_setpoint_schedule(self.id, schedule_df)

    def get_schedule(self, start_date, end_date) -> pd.DataFrame:
        return get_setpoint_schedule(self.id, start_date, end_date)



from datetime import timezone
from typing import Dict, List, Optional, Union

import pandas as pd

from .api import TIMEZONE
from .component import Component
from .data import load_model_data
from .helpers import (
    now,
    sanitise_datetime_input,
    time_resolution_aliases,
    timedelta,
    utc,
)
from .logging import Logging


def populate_units(unit_type, units_data):
    units = []
    # Loop over units; create and populate Unit instances
    for unit_i in units_data:
        units.append(Unit(unit_i))

    return units


class Unit(object):
    """Defines a unit in the PreHEAT sense"""

    def __init__(self, unit_type: str, unit_data: Dict):
        # Type
        self.unit_type: str = unit_type
        self.unit_subtype: Optional[str] = unit_data.pop("type", None)

        # Identifier of the unit
        self.id = unit_data.pop("id", None)
        # Name of the unit
        self.name: str = unit_data.pop("name", "{}_{}".format(self.unit_type, self.id))

        # List of components in the unit (PreHEAT_API.Component)
        self.components: List[Component] = self.__populate_components(unit_data)

        # Time series cache
        self.data: pd.DataFrame = pd.DataFrame()

        # State cache
        self._state: pd.DataFrame = pd.DataFrame()

    def __repr__(self) -> str:
        return "{0}({1})".format(self.unit_type, self.name)

    def __populate_components(self, unit_data: Dict) -> List[Component]:
        components = []
        try:
            for key, value in unit_data.items():
                # Component properties: a dict w. key 'cid'
                if isinstance(value, dict) and "cid" in value:
                    # Let 'name' be PreHEAT name, and tag be BACNET/source name
                    value["tag"] = value["name"]
                    value["name"] = key
                    # Add PreHEAT name as description
                    components.append(Component(value))

        except TypeError:
            pass

        return components

    def load_data(
        self,
        start_date,
        end_date,
        time_resolution: str = "minute",
        components: Union[List, None] = None,
        **kwargs,
    ) -> None:
        self._warn_if_data_is_loaded()
        self.data = load_model_data(
            self.get_all_component_ids(components=components),
            start_date,
            end_date,
            time_resolution,
        )
        self._ensure_continuity_of_data(time_resolution)

    def _ensure_continuity_of_data(self, time_resolution: str) -> None:
        if self.data.empty or (time_resolution == "raw"):
            return

        time_alias = time_resolution_aliases(time_resolution)

        if time_resolution in ["day", "week", "month", "year"]:
            # Reindexing to local time prior to frequency conversion
            self.data.index = self.data.index.tz_convert(TIMEZONE)
            self.data = self.data.asfreq(freq=time_alias)
            self.data.index = self.data.index.tz_convert(utc)
        else:
            self.data = self.data.asfreq(freq=time_alias)

    def _warn_if_data_is_loaded(self):
        if self.data.empty is False:
            Logging().warning(
                RuntimeWarning(
                    f"Data was already present in unit (id={self.id}, name={self.name}, type={self.unit_type})"
                )
            )

    def clear_data(self, **kwargs) -> None:
        self.data = self.data[0:0]

    def cquery(self, name: str):
        return [component for component in self.components if component.name == name][0]

    def _select_components(self, components: Union[List, None] = None) -> List:
        if components is None:
            return self.components
        else:
            if isinstance(components, str):
                components = [components]
            return [
                component
                for component in self.components
                if component.name in components
            ]

    def get_all_component_cids(
        self, prefix: bool = False, components: Union[List, None] = None
    ):
        prefix = "{}.".format(self.name) if prefix else ""
        comps = self._select_components(components=components)
        return {str(c.cid): prefix + c.name for c in comps}

    def get_all_component_ids(
        self, prefix: bool = False, components: Union[List, None] = None
    ) -> Dict:
        prefix = "{}.".format(self.name) if prefix else ""
        comps = self._select_components(components=components)
        return {str(c.id): prefix + c.name for c in comps}

    def get_all_component_details(
        self, prefix: bool = False, components: Union[List, None] = None
    ) -> List:
        prefix = "{}.".format(self.name) if prefix else ""
        comps = self._select_components(components=components)

        return [
            {
                "id": str(c.id),
                "cid": str(c.cid),
                "name": prefix + c.name,
                "stdUnitDevisor": c.std_unit_devisor,
                "stdUnit": c.std_unit,
            }
            for c in comps
        ]

    def get_component(self, name: str) -> Component:
        try:
            return [
                component for component in self.components if component.name == name
            ][0]
        except:
            raise Exception(
                "The component ({}) does not exist in the unit ({}).".format(
                    name, self.name
                )
            )

    def has_component(self, component_name: str) -> bool:
        for comp_i in self.components:
            if comp_i.name == component_name:
                return True
        return False

    def has_data(self, component=None, check_not_null: bool = True):
        if component is None:
            return len(self.data) > 0
        elif self.has_component(component) is False:
            return False
        elif self.data.empty is True:
            return False
        elif component in self.data.columns and self.data.shape[0] > 0:
            return self.data[component].notnull().any() if check_not_null else True
        else:
            return False

    # Methods to manage unit state
    def load_state(self, seconds_back: int = 300, t_now=None) -> None:
        t_now = now() if t_now is None else sanitise_datetime_input(t_now)
        # Check if recent enough to expect raw data
        if t_now > now() - timedelta(days=7):
            resolution = "raw"
        else:
            resolution = "minute"
        self._state = load_model_data(
            self.get_all_component_ids(),
            t_now - timedelta(seconds=seconds_back),
            t_now,
            resolution,
        )

    def clear_state(self, **kwargs) -> None:
        self._state = pd.DataFrame()

    def get_state(
        self,
        update: bool = False,
        estimate: str = "last",
        seconds_back: int = 300,
        t_now=None,
    ) -> pd.Series:
        if update is True:
            self.load_state(seconds_back=seconds_back, t_now=t_now)

        if (self._state is None) or (self._state.empty):
            state = pd.Series([], dtype=float)
        elif estimate == "last":
            state = self._state.ffill().iloc[-1, :]
        elif estimate == "mean":
            state = self._state.mean()
        elif estimate == "median":
            state = self._state.median()
        else:
            raise Exception(f"Invalid value for estimate ({estimate})")

        return state



import pandas as pd
from requests import HTTPError
from requests.models import Response

from .api import TIMEZONE, api_get, api_put
from .helpers import sanitise_datetime_input


class InvalidSchedule(BaseException):
    pass


def send_setpoint_schedule(controlunit_id, schedule_df: pd.DataFrame) -> Response:

    path = f"controlunit/{controlunit_id}/setpoint"

    # Build payload from DataFrame
    payload = {"schedule": []}
    if "operation" not in schedule_df.keys():
        schedule_df["operation"] = len(schedule_df) * ["NORMAL"]

    for index, row in schedule_df.iterrows():
        payload["schedule"].append(
            {
                "startTime": index.isoformat(),
                "value": row["value"],
                "operation": row["operation"],
            }
        )

    try:
        # PUT request
        return api_put(path, json_payload=payload)
    except HTTPError as e:
        extra_info = (
            f"""Error when writing the schedule: \n {str(payload["schedule"])}"""
        )
        raise InvalidSchedule(extra_info) from e


def get_setpoint_schedule(controlunit_id, start_date, end_date) -> pd.DataFrame:
    start_date = sanitise_datetime_input(start_date)
    end_date = sanitise_datetime_input(end_date)

    path = f"controlunit/{controlunit_id}/setpoint"

    payload = {
        "start_time": start_date.isoformat(),
        "end_time": end_date.isoformat(),
    }

    raw = api_get(path, out="json", payload=payload)
    df = pd.DataFrame(raw)

    if df.empty:
        return df

    df["startTime"] = pd.to_datetime(df["startTime"], utc=True)
    return df.set_index("startTime")



import pandas as pd

from preheat_open import test


class TestBuilding(test.PreheatTest):
    def test_building(self, building):
        assert building is not None
        assert building.weather is not None
        assert building.units is not None
        assert len(building.units) > 0
        assert building.units["heating"] is not None

    def test_query_units(self, building, unit_id):
        assert building.query_units(unit_id=unit_id) is not None
        assert building.query_units("heating") is not None
        assert building.query_units("?h") is not None
        assert building.query_units("?ting") is not None
        assert building.qu(name="heating_primary") is not None

        assert building.query_units("definitely not there") is not None
        assert len(building.query_units("definitely not there")) == 0
        assert building.qu(name="definitely not there") is None
        assert building.qu(unit_id="definitely not there") is None

    def test_get_unit_graph(self, building):
        assert building.get_unit_graph() is not None

    def test_clear_data(self, building_with_data):
        from copy import deepcopy

        building_new = deepcopy(building_with_data)
        building_new.clear_data()
        assert building_new.weather is not None
        assert building_new.weather.data is not None
        assert building_new.weather.data.empty

    def test_load_data(self, building_with_data, short_period):
        # Testing standard data loading
        assert building_with_data is not None
        assert building_with_data.weather is not None
        assert building_with_data.weather.data is not None
        assert not building_with_data.weather.data.empty

        # Testing selective data loading
        building_with_data.clear_data()
        comps = {
            "weather": ["Temperature", "Humidity"],
            "main": None,
            "secondaries": ["supplyT"],
        }
        building_with_data.load_data(*short_period, components=comps)

        df_weather = building_with_data.weather.data
        df_main = building_with_data.query_units("main")[0].data
        df_secondary = building_with_data.query_units("secondaries")[0].data

        assert df_weather.empty is False
        assert "Temperature" in df_weather
        assert "Humidity" in df_weather
        assert "WindSpeed" not in df_weather

        assert df_main.empty is False

        assert df_secondary.empty is False
        assert "supplyT" in df_secondary
        assert "returnT" not in df_secondary

        assert building_with_data.query_units("indoorClimate")[0].data.empty is True
        assert building_with_data.query_units("heating")[0].data.empty is True

    def test_get_sub_zones(self, building_with_data):
        assert building_with_data.get_zones([]) == []
        assert len(building_with_data.get_zones([1])) == 0
        assert len(building_with_data.get_zones([4830])) == 1
        assert len(building_with_data.get_zones([4830, 4834])) == 2


def test_available_buildings():
    from preheat_open.building import available_buildings

    df = available_buildings()
    assert isinstance(df, pd.DataFrame)



from preheat_open import test


class TestBuildingUnit(test.PreheatTest):
    def test_class(self, unit_with_data):
        assert unit_with_data.is_shared() is False

    def test_load_data(self, unit_with_data):
        assert unit_with_data is not None
        assert unit_with_data.data is not None
        assert not unit_with_data.data.empty
        assert unit_with_data.has_data()
        print(unit_with_data.data)
        assert unit_with_data.has_data("supplyT")

        # TODO add a test of load_data(self, *, load_children=True)
        assert True is True

    def test_clear_data(self, unit_with_data):
        unit_with_data.clear_data()
        assert unit_with_data.data is not None
        assert unit_with_data.data.empty

        unit_with_data.clear_data(clear_children=True)
        assert unit_with_data.data is not None
        assert unit_with_data.data.empty

    def test_get_zones(self, building_with_data):
        from preheat_open.zone import Zone

        zones_ok = building_with_data.qu("indoorClimate").get_zones()
        zones_empty = building_with_data.qu("custom").get_zones()
        assert len(zones_ok) == 1
        assert isinstance(zones_ok[0], Zone)
        assert len(zones_empty) == 0

    def test_query_units(self, building_with_data):
        m = building_with_data.qu("main", unit_id=15312)
        sec = building_with_data.qu("secondaries", unit_id=15401)
        c1 = sec.qu("control", name="dummy_heating_controller")

        assert sec.unit_subtype == "MIXING_LOOP"
        assert m.unit_subtype is None

        assert m.id == sec.qpu("main").id
        assert m.id == c1.qpu("main").id
        assert c1.id == m.qu("control", name="dummy_heating_controller").id
        assert (
            c1.id
            == building_with_data.qu("control", name="dummy_heating_controller").id
        )
